/**
 * Copyright Jan Newmarch
 * BSD License
 */

import java.nio.ByteBuffer;
import java.net.DatagramPacket; 
import java.net.DatagramSocket;
import java.net.SocketTimeoutException;
import java.net.InetAddress;
import java.util.Vector;
import java.util.Arrays;

public class Connection {

    // Grrrr, no unsigned byte type
    // 0xaa -> -86, 0xa0 -> -96
    private static byte first[] = new byte[]{-86, 0x21, -96, 0x10};
    //protected static byte second[] = new byte[]{0x20, 0x15, 0x04, 0x63};
    private static byte third[] = new byte[]{0x00, 0x21, -96, 0x10};
    private static byte last[] = new byte[17]; // zero's padding to 33 bytes

    private static DatagramSocket socket;

    static {
	try {
	    socket = new DatagramSocket(8866);
	    socket.setSoTimeout(1200); // has SocketTimeoutException
	} catch(Exception e) {
	    System.out.println("Can't create socket");
	    System.exit(1);
	}
	System.out.println("Connection inited");
    }

    public static void discardBufferedPackets() {
	while (true) {
	    try {
		getPacket();
		System.out.println("Discarding pkt");
	    } catch(SocketTimeoutException e) {
		System.out.println("Timeout from discarding pkts");
		return;
	    }
	}
    }
    
    public static boolean setState(byte ID[], byte type, byte val, byte check) {
	ByteBuffer buf = ByteBuffer.allocate(33);
	byte data[];
	DatagramPacket nextPkt = null;

	discardBufferedPackets();
	
	buf.put(first);
	buf.put(check);
	//buf.put(second);
	buf.put(ID);
	buf.put(type);
	buf.put(val);
	buf.put(third);
	buf.put(last);

	data = buf.array();
	
	while (true) {
	    sendPacket("255.255.255.255", data);
	    //sendPacket("192.168.1.255", data);
	    System.out.println("Sent data pkt");

	    try {
		// this should be the packet we just sent
		nextPkt = getPacket();
		
		// this should be the reply if it worked
		nextPkt = getPacket();
	    } catch(SocketTimeoutException se) {
		// ignore
	    }
	    if (nextPkt != null && nextPkt.getLength() == 33 &&
		nextPkt.getData()[2] == 0x0a) {
		// should check reply from right host
		System.out.println("Ending send");
		break;
	    }
	    // it didn't work so try again
	    
	    try {
		Thread.sleep(250);
	    } catch(Exception e) {
		// ignore
	    }
	    
	}
	return true;
    }

    private static void sendPacket(String IPaddr, byte data[]) {
	try {
	    InetAddress addr = InetAddress.getByName(IPaddr);
	    DatagramPacket pkt = new DatagramPacket(data, data.length, addr, 8866);
	    socket.send(pkt);
	} catch(Exception e) {
	    System.out.println("Failed to send packet " + e.toString());
	}
    }

    private static DatagramPacket getPacket() throws SocketTimeoutException {
	byte buffer[] = new byte[2048];

	DatagramPacket packet = new DatagramPacket(buffer, buffer.length);
	try {
	    socket.receive(packet);
	} catch(SocketTimeoutException se) {
	    throw(se);
	} catch(Exception e) {
	    System.out.println("Socket read failed");
	    return null;
	}
	System.out.println("Received pkt from " + packet.getAddress() +
			   " length " + packet.getLength());
	return packet;
    }

    public static Switch[] discoverSwitches() {
	Vector<Switch> switches = new Vector<Switch>();
	// Vector<DatagramPacket> pkts = new Vector<DatagramPacket>();
	
	byte discoverData[] = new byte[] {
	    /*
	    0xaa, 0x21, 0xa0, 0x09, 0x59, 0x71, 0x2f, 0x34,
	    0x05, 0x6c, 0x00, 0x00, 0x0b, 0x00, 0xa0, 0x11,
	    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	    0x00
	    */
	    -86, 0x21, -96, 0x09, 0x59, 0x71, 0x2f, 0x34,
	    0x05, 0x6c, 0x00, 0x00, 0x0b, 0x00, -96, 0x11,
	    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	    0x00

	};

	Vector<InetAddress> seenAddresses = new Vector<InetAddress>();

	// search for 10 seconds
	long now = System.currentTimeMillis();
	while (System.currentTimeMillis() - now <= 10000) {
	    //for (int n = 0; n < 10; n++) {
	    Connection.sendPacket("255.255.255.255", discoverData);
	    while (true) {
		try {
		    DatagramPacket pkt = getPacket();
		    System.out.println("Got packet " + pkt);
		    byte[] data = pkt.getData();
		    if (pkt.getLength() == 33 &&
			data[2] == 0x0a) {
			// should be a switch declaring itself
			boolean seenAddr = false;
			for (InetAddress addr: seenAddresses) {
			    if (addr.equals(pkt.getAddress())) {
				seenAddr = true;
				break;
			    }
			}
			if (!seenAddr) {
			    seenAddresses.add(pkt.getAddress());
			    System.out.println("New switch from " + pkt.getAddress());
			    byte switchBytes[] = Arrays.copyOfRange(data,
								    5,
								    12);
			    System.out.println("with bytes " + Arrays.toString(switchBytes));
			    Switch newSwitch = newSwitch(switchBytes);
			    if (newSwitch != null)
				switches.add(newSwitch);
			}
		    }
				
		} catch(SocketTimeoutException se) {
		    break;
		}
	    }
	}
	return switches.toArray(new Switch[switches.size()]);
    }

    private static Switch newSwitch(byte[] bytes) {
	byte[] ID = Arrays.copyOfRange(bytes, 0, 5);
	byte kind = bytes[5];
	byte value = bytes[6];

	switch (kind) {
	case 0x01: return new OneGangSwitch(new String[]{"single"},
					    ID, value);
	case 0x02: return new TwoGangSwitch(new String[]{"left", "right"},
					    ID, value);
	case 0x03: return null;
	case 0x04: return new DimmableSwitch(new String[]{"dimmable"},
					    ID, value);
	}
	return null;
    }
}
